from dataclasses import dataclass


def combined_parameters(user_params: dict, default_params: dict):
    combined = {k: user_params.get(k, default_params[k]) for k in default_params}
    return combined


@dataclass
class Tank:
    from .pack import Pack

    packs: tuple[Pack, ...]

    @staticmethod
    def from_calibration(**params):
        from scipp import vector
        from scipp.spatial import rotations_from_rotvecs
        from .pack import Pack
        from .parameters import known_pack_params, tube_xy_displacement_to_quaternion
        combined = combined_parameters(params, known_pack_params())

        sample = combined['sample']  # the origin of the detector positions
        # (,), (tube,), (pack,)  or (pack, tube)
        detector_vector = vector([1, 0, 0]) * combined['sample_detector_distance']

        length = combined['detector_length'].to(unit='m')
        resistance = combined['resistance']
        resistivity = combined['resistivity']

        # (tube,)
        tube_rotations = rotations_from_rotvecs(combined['tube_angles'] * vector([0, 0, 1]))
        # (pack, )
        pack_rotations = rotations_from_rotvecs(combined['pack_angles'] * vector([0, 0, 1]))
        # (pack, tube)
        tube_positions = sample + pack_rotations * tube_rotations * detector_vector
        # (,), or (pack, tube)
        tube_orient = tube_xy_displacement_to_quaternion(length, combined['detector_orient'].to(unit='m'))
        # (pack, tube)
        tube_orient = pack_rotations * tube_rotations * tube_orient

        # we need to define (position: Variable, length: Variable, **params) for each pack
        n_packs = tube_positions.sizes['pack']

        orient_per, resistance_per, resistivity_per, length_per = ['pack' in x.dims for x in (tube_orient, resistance, resistivity, length)]

        pack_list = []
        for n in range(n_packs):
            pp = dict(sample=sample, orient=tube_orient['pack', n] if orient_per else tube_orient,
                      resistivity=resistivity['pack', n] if resistivity_per else resistivity,
                      resistance=resistance['pack', n] if resistance_per else resistance)
            pp['radius'] = combined['detector_radius']
            p = Pack.from_calibration(tube_positions['pack', n], length['pack', n] if length_per else length, **pp)
            pack_list.append(p)

        return Tank(tuple(pack_list))

    def to_mccode(self, assembler, sample, **kwargs):
        from ..mccode import ensure_user_var
        ensure_user_var(assembler, 'int', 'flag', 'Indicate detection in a monitor')

        group = 'packs_group'
        for index, pack in enumerate(self.packs):
            name = f'pack_{1+index:02d}'
            extend = "flag = (SCATTERED) ? 1 : 0;"
            pack.to_mccode(
                assembler,
                sample,
                name=name,
                extend=extend,
                group=group,
                first_wire_index=index * 24,
                **kwargs
            )



from typing import Optional, Union

from scipp import Variable
from mccode_antlr.instr import Instance
from mccode_antlr.common.parameters import InstrumentParameter
from mccode_antlr.assembler import Assembler
from .component import Component


ESS_SOURCE_DURATION = Variable(values=2.857e-3, unit='s', dims=None)



class Source(Component):
    pass


class ESSource(Source):
    """Representation of the ESS Butterfly source

    https://github.com/mccode-dev/McCode/blob/main/mcstas-comps/sources/ESS_butterfly.comp
    """
    sector: str
    beamline: int
    height: Variable
    cold_frac: float
    focus_distance: Optional[Variable]
    focus_width: Optional[Variable]
    focus_height: Optional[Variable]
    cold_performance: float
    thermal_performance: float
    wavelength_minimum: Optional[Union[Variable, InstrumentParameter]]
    wavelength_maximum: Optional[Union[Variable, InstrumentParameter]]
    latest_emission_time: Optional[Variable]
    n_pulses: Optional[int]
    accelerator_power: Optional[Variable]

    @classmethod
    def from_calibration(cls, cal: dict):
        from scipp import vector as v, scalar as s
        from scipp.spatial import rotations_from_rotvecs as r
        from niess.io.mccode import reconstitute_instrument_parameter as rip
        name = cal.get('name', 'ESS_source')
        position = cal.get('position', v([0, 0, 0.], unit='m'))
        orientation = cal.get('orientation', r(v([0, 0, 0.], unit='rad')))

        sector = cal.get('sector', 'W')
        beamline = cal.get('beamline', 4)
        height = cal.get('height', s(3.0, unit='cm'))
        cold_frac = cal.get('cold_fraction', 0.5)
        focus_distance = cal.get('focus_distance', None)
        focus_width = cal.get('focus_width', None)
        focus_height = cal.get('focus_height', None)
        cold_performance = cal.get('cold_performance', 1.0)
        thermal_performance = cal.get('thermal_performance', 1.0)
        wavelength_minimum = rip(cal.get('wavelength_minimum', None), (Variable,))
        wavelength_maximum = rip(cal.get('wavelength_maximum', None), (Variable,))
        latest_emission_time = cal.get('latest_emission_time', None)
        n_pulses = cal.get('n_pulses', None)
        accelerator_power = cal.get('accelerator_power', None)

        return cls(
            name=name,
            position=position,
            orientation=orientation,
            sector=sector,
            beamline=beamline,
            height=height,
            cold_frac=cold_frac,
            focus_distance=focus_distance,
            focus_width=focus_width,
            focus_height=focus_height,
            cold_performance=cold_performance,
            thermal_performance=thermal_performance,
            wavelength_minimum=wavelength_minimum,
            wavelength_maximum=wavelength_maximum,
            latest_emission_time=latest_emission_time,
            n_pulses=n_pulses,
            accelerator_power=accelerator_power
        )

    def __mccode__(self) -> tuple[str, dict]:
        from ..utilities import variable_value_or_parameter as value_or
        pars = {
            'sector': '"' + self.sector.strip('"') + '"',
            'beamline': self.beamline,
            'yheight': self.height.to(unit='m').value,
            'cold_frac': self.cold_frac,
            'c_performance': self.cold_performance,
            't_performance': self.thermal_performance,
        }
        if all(x is not None for x in (self.focus_width, self.focus_height, self.focus_distance)):
            pars['dist'] = self.focus_distance.to(unit='m').value
            pars['focus_xw'] = self.focus_width.to(unit='m').value
            pars['focus_yh'] = self.focus_height.to(unit='m').value
        if all(x is not None for x in (self.wavelength_minimum, self.wavelength_maximum)):
            pars['Lmin'] = value_or(self.wavelength_minimum, 'angstrom')
            pars['Lmax'] = value_or(self.wavelength_maximum, 'angstrom')
        if self.latest_emission_time is not None:
            multiplier = self.latest_emission_time.to(unit=ESS_SOURCE_DURATION.unit) / ESS_SOURCE_DURATION
            pars['tmax_multiplier'] = multiplier.value
        if self.n_pulses is not None:
            pars['n_pulses'] = self.n_pulses
        if self.accelerator_power is not None:
            pars['acc_power'] = self.accelerator_power.to(unit='MW').value

        return 'ESS_butterfly', pars

    def to_mccode(
            self, assembler: Assembler,
            at: Instance | str | None = None, rotate: Instance | str | None = None,
    ):
        from ..mccode import ensure_runtime_parameter
        for field in self.fields():
            p = getattr(self, field)
            if isinstance(p, InstrumentParameter):
                ensure_runtime_parameter(assembler, p)
        return super().to_mccode(assembler, at, rotate)




from __future__ import annotations
from pathlib import Path
import msgspec
from niess.io.utils import encode_hook, decode_hook, Model, from_model


def to_json(obj) -> bytes:
    enc_hook = encode_hook(msgspec.json.Encoder())
    encoder = msgspec.json.Encoder(enc_hook=enc_hook)
    return encoder.encode(Model.from_value(obj, encoder=encoder))


def from_json(msg: bytes):
    from niess.io.utils import MODEL_DECODE
    decoded = msgspec.json.decode(msg, type=Model)

    dec_hook = decode_hook(msgspec.json.Decoder())
    ## FIXME The first (typed) version causes the TypeError
    ##       "unsupported operand type(s) for |: 'classmethod' and 'classmethod'"
    ##       _probably_ due to mccode_antlr.common.Expr using Union[...]
    ##       _without_ specifying to msgspec that its Structs should store type
    ##       information.
    # if decoded.name in MODEL_DECODE:
    #     decoder = msgspec.json.Decoder(MODEL_DECODE[decoded.name], dec_hook=dec_hook)
    # else:
    #     decoder = msgspec.json.Decoder(dec_hook=dec_hook)
    decoder = msgspec.json.Decoder(dec_hook=dec_hook)
    return from_model(decoder, decoded)


def save_json(obj, filename: str | Path) -> None:
    with open(filename, "wb") as f:
        f.write(to_json(obj))


def load_json(filename: str | Path):
    with open(filename, "rb") as f:
        return from_json(f.read())


import scipp as sc


def scipp_round_trip(variable):
    import msgspec
    from niess.io.scipp import ScippModel, from_scipp_model
    encoder = msgspec.json.Encoder()
    modeled = ScippModel.from_value(variable, encoder=encoder)
    json = encoder.encode(modeled)

    decoded = msgspec.json.decode(json, type=ScippModel)
    assert modeled == decoded
    return from_scipp_model(msgspec.json.Decoder(), decoded)


def niess_round_trip(value):
    import msgspec
    from niess.io.utils import Model, from_model, decode_hook, MODEL_DECODE
    from niess.io.utils import encode_hook
    encoder = msgspec.json.Encoder(enc_hook=encode_hook(msgspec.json.Encoder()))
    modeled = Model.from_value(value, encoder=encoder)
    json = encoder.encode(modeled)
    decoded = msgspec.json.decode(json, type=Model)
    assert modeled == decoded
    dec_hook = decode_hook(msgspec.json.Decoder())
    decoder = msgspec.json.Decoder(dec_hook=dec_hook)
    return from_model(decoder, decoded)


def test_scipp_scalar():
    p = sc.scalar(3.14159, unit='rad')
    assert p == scipp_round_trip(p)

def test_scipp_vector():
    p = sc.vector([1,2,3.], unit='m')
    assert p == scipp_round_trip(p)


def test_scipp_rotation():
    q = sc.spatial.rotation(value=[0,0,0,1.])
    assert q == scipp_round_trip(q)


def test_scipp_array():
    from numpy.random import random
    a = sc.array(values=random((10,20)), variances=random((10,20)), dims=['a', 'b'], unit='m')
    res = scipp_round_trip(a)
    assert sc.allclose(res, a)


def test_mccode_types():
    from mccode_antlr.common.parameters import InstrumentParameter
    par = InstrumentParameter.parse('int named/"K" = -1')
    print(par)
    assert par.name == 'named'
    assert par.unit == '"K"'
    assert par.value.value == -1
    ret = niess_round_trip(par)
    assert par == ret


def test_fake_niess_component():
    from niess.components import He3Monitor
    from pytest import raises
    from msgspec import ValidationError
    b = He3Monitor(
        name='mon1',
        position=[1,2,3],
        orientation=[0,0,0,1],
        radius=10,
        length=20,
        pressure=1
    )
    # If the decoder used the typed variant, this would fail ... but other
    # types wouldn't decode since they include McCode_antlr objects ...
    # # Fails due to the incorrect data types provided
    # with raises(ValidationError):
    #     niess_round_trip(b)
    assert b == niess_round_trip(b)


def test_niess_component():
    from niess.components import He3Monitor
    import scipp as sc
    b = He3Monitor(
        name='mon2',
        position=sc.vector([1, 2, 3], unit='m'),
        orientation=sc.spatial.rotation(value=[0, 0, 0, 1]),
        radius=sc.scalar(10., unit='mm'),
        length=sc.scalar(20, unit='mm'),
        pressure=sc.scalar('1', unit='atm')
    )
    assert b == niess_round_trip(b)


def test_niess_source():
    from niess.components import ESSource
    source = ESSource.from_calibration(
        dict(wavelength_minimum='double minimum/"angstrom"=1.0',)
    )
    assert source == niess_round_trip(source)


def test_section():
    from niess.components import Section
    section = Section()
    returned = niess_round_trip(section)
    assert section == returned


def test_bifrost_primary():
    from niess.bifrost.parameters import primary_parameters
    from niess.bifrost import Primary
    primary = Primary.from_calibration(primary_parameters())
    assert primary == primary
    returned = niess_round_trip(primary)
    assert primary == returned


def test_bifrost_tank():
    from niess.bifrost.parameters import tank_parameters
    from niess.bifrost import Tank
    tank = Tank.from_calibration(tank_parameters())
    assert tank == tank
    returned = niess_round_trip(tank)
    assert tank == returned


from typing import Any, Type
import msgspec

from mccode_antlr.io.utils import MODEL_ENC

MCCODE_MODEL_ENCODE = {k: f'mccode.{v}' for k, v in MODEL_ENC.items()}
MCCODE_MODEL_DECODE = {v: k for k, v in MCCODE_MODEL_ENCODE.items()}


class McCodeModel(msgspec.Struct):
    name: str
    obj: msgspec.Raw

    @classmethod
    def from_value(cls, obj: Any, encoder=None):
        if encoder is None:
            raise ValueError("And encoder must be provided")
        model_type = MCCODE_MODEL_ENCODE[type(obj)]
        if model_type in MCCODE_MODEL_DECODE and hasattr(obj, 'to_dict'):
            obj = obj.to_dict()
        return cls(model_type, msgspec.Raw(encoder.encode(obj)))


def reconstitute_instrument_parameter(a: Any, others: tuple[Type,...]):
    from mccode_antlr.common import InstrumentParameter
    if a is None or isinstance(a, InstrumentParameter) or any(isinstance(a, x) for x in others):
        return a
    if isinstance(a, str):
        return InstrumentParameter.parse(a)
    if isinstance(a, dict):
        return InstrumentParameter.from_dict(a)
    raise ValueError(f"Unknown relationship of type {type(a)} to InstrumentParameter")


from niess.components.component import Base

class Analyzer(Base):
    from mccode_antlr.assembler import Assembler
    from ..components import Crystal
    from scipp import Variable

    blades: tuple[Crystal, ...]  # 7-9 blades

    @classmethod
    def from_dict(cls, data):
        from ..components import Crystal
        blades = data['blades']
        if not hasattr(blades, '__len__') or (len(blades) != 7 and len(blades) != 9):
            raise ValueError('Blades must have 7 or 9 elements')
        blades = tuple(b if isinstance(b, Crystal) else Crystal.from_dict(b) for b in blades)
        return cls(blades)

    @property
    def central_blade(self):
        return self.blades[len(self.blades) >> 1]

    @property
    def count(self):
        return len(self.blades)

    @staticmethod
    def from_calibration(position: Variable, focus: Variable, tau: Variable, **params):
        from scipp import scalar, vector
        from scipp.spatial import rotation
        from ..spatial import is_scipp_vector
        from .rowland import rowland_blades
        from ..components import Crystal
        map(lambda x: is_scipp_vector(*x), ((position, 'position'), (focus, 'focus'), (tau, 'tau')))
        count = params.get('blade_count', scalar(9))  # most analyzers have 9 blades
        shape = params.get('shape', vector([10., 200., 2.], unit='mm'))
        orient = params.get('orient', None)
        orient = rotation(value=[0, 0, 0, 1.]) if orient is None else orient
        mosaic = params.get('mosaic', scalar(40., unit='arcminutes'))
        # qin_coverage = params.get('qin_coverage', params.get('coverage', scalar(0.1, unit='1/angstrom')))
        coverage = params.get('coverage', scalar(2.0, unit='degree'))
        source = params.get('source', params.get('sample_position', vector([0, 0, 0], unit='m')))
        gap = params.get('gap', None)
        #
        # Use the Rowland geometry to define each blade position & normal direction
        positions, taus = rowland_blades(source, position, focus, coverage, shape.fields.x, count.value, tau, gap)

        blades = [Crystal(p, t, shape, orient, mosaic) for p, t in zip(positions, taus)]
        return Analyzer(tuple(blades))

    def triangulate(self, unit=None):
        from ..spatial import combine_triangulations
        vts = [blade.triangulate(unit=unit) for blade in self.blades]
        return combine_triangulations(vts)

    def bounding_box(self, basis: Variable, unit=None):
        from scipp import concat
        from ..spatial import combine_bounding_boxes
        boxes = concat(tuple(b.bounding_box(basis, unit) for b in self.blades), dim='blades')
        return combine_bounding_boxes(boxes)

    def coverage(self, sample: Variable, unit=None):
        from scipp import norm, dot, cross, max, min, atan2, vector, concat
        unit = unit or 'radian'
        # Define a pseudo McStas coordinate system (requiring y is mostly vertical)
        z = (self.central_blade.position - sample)
        dist = norm(z)
        z = z / dist
        y = cross(cross(z, vector([0, 0, 1.0])), z)
        y = y / norm(y)
        x = cross(y, z)  # should have length 1 but normalized in bounding_box
        basis = concat((x, y), dim='basis')
        box = self.bounding_box(basis, unit=dist.unit)
        lengths = box['limits', 1] - box['limits', 0]
        return tuple(2 * atan2(y=e/2, x=dist).to(unit=unit) for e in lengths)

    def sample_space_angle(self, sample: Variable):
        from scipp import dot, atan2, vector
        z = (self.central_blade.position - sample)
        sample_space_x = vector([1, 0, 0])
        sample_space_y = vector([0, 1, 0])
        return atan2(y=dot(sample_space_y, z), x=dot(sample_space_x, z)).to(unit='radian')

    def rtp_parameters(self, sample: Variable, oop: Variable):
        from scipp import concat
        p0 = self.central_blade.position
        # exploit that for x in zip returns first all the first elements, then all the second elements, etc.
        x, y, a = [concat(x, dim='blades') for x in zip(*[b.rtp_parameters(sample, p0, oop) for b in self.blades])]
        return x, y, a

    def mcstas_parameters(self, sample: Variable, source: str, sink: str) -> dict:
        from mccode_antlr.instr import Instance
        from ..spatial import is_scipp_vector
        is_scipp_vector(sample, 'sample')
        source = source.name if isinstance(source, Instance) else source
        sink = sink.name if isinstance(sink, Instance) else sink
        if not isinstance(source, str) or not isinstance(sink, str):
            raise ValueError(f'The source and sink are expected to be str values not {type(source)} and {type(sink)}')

        perp_q, perp_plane, parallel_q = self.central_blade.shape.to(unit='m').value
        hor_cov, ver_cov = self.coverage(sample)
        params = dict(
            NH=self.count,
            zwidth=perp_q,
            yheight=perp_plane,
            mosaic=self.central_blade.mosaic.to(unit='arcminute').value,
            DM=3.355,
            gap=0.002,
            show_construction='showconstruction',
            angle_h=ver_cov.to(unit='degree').value,
            source=f'"{source}"',
            sink=f'"{sink}"'
        )
        return params

    def to_mccode(self, assembler: Assembler, source: str, relative: str, sink: str, theta: float, name: str,
                  when: str = None, extend: str = None, origin: Variable = None):
        mono = assembler.component(name, 'Monochromator_Rowland',
                                   at=((0, 0, 0), relative), rotate=((0, theta, 0), relative))
        mono.set_parameters(**self.mcstas_parameters(origin, source, sink))
        mono.WHEN(when)
        mono.EXTEND(extend)

